@@ -39,6 +39,8 @@
 #include "nsIDOMHTMLInputElement.h"
 #include "nsIDOMNSHTMLInputElement.h"
 #include "nsITextControlElement.h"
+#include "nsIFileControlElement.h"
+#include "nsIDOMNSEditableElement.h"
 #include "nsIRadioControlElement.h"
 #include "nsIRadioVisitor.h"
 #include "nsIPhonetic.h"
@@ -136,7 +138,9 @@ class nsHTMLInputElement : public nsGenericHTMLFormElement,
                            public nsIDOMNSHTMLInputElement,
                            public nsITextControlElement,
                            public nsIRadioControlElement,
-                           public nsIPhonetic
+                           public nsIPhonetic,
+                           public nsIDOMNSEditableElement,
+                           public nsIFileControlElement
 {
 public:
   nsHTMLInputElement(nsINodeInfo *aNodeInfo, PRBool aFromParser);
@@ -163,8 +167,11 @@ public:
   // nsIPhonetic
   NS_DECL_NSIPHONETIC
 
+  // nsIDOMNSEditableElement
+  NS_FORWARD_NSIDOMNSEDITABLEELEMENT(nsGenericHTMLElement::)
+
   // Overriden nsIFormControl methods
-  NS_IMETHOD_(PRInt32) GetType() { return mType; }
+  NS_IMETHOD_(PRInt32) GetType() const { return mType; }
   NS_IMETHOD Reset();
   NS_IMETHOD SubmitNamesValues(nsIFormSubmission* aFormSubmission,
                                nsIContent* aSubmitElement);
@@ -206,8 +213,6 @@ public:
 
     nsresult rv = nsGenericHTMLFormElement::SetAttr(aNameSpaceID, aName,
                                                     aPrefix, aValue, aNotify);
-
-    AfterSetAttr(aNameSpaceID, aName, &aValue, aNotify);
     return rv;
   }
 
@@ -218,8 +223,6 @@ public:
 
     nsresult rv = nsGenericHTMLFormElement::UnsetAttr(aNameSpaceID, aAttribute,
                                                       aNotify);
-
-    AfterSetAttr(aNameSpaceID, aAttribute, nsnull, aNotify);
     return rv;
   }
 
@@ -230,6 +233,10 @@ public:
   // nsITextControlElement
   NS_IMETHOD TakeTextFrameValue(const nsAString& aValue);
   NS_IMETHOD SetValueChanged(PRBool aValueChanged);
+  
+  // nsIFileControlElement
+  virtual void GetFileName(nsAString& aFileName);
+  virtual void SetFileName(const nsAString& aFileName, PRBool aUpdateFrame);
 
   // nsIRadioControlElement
   NS_IMETHOD RadioSetChecked(PRBool aNotify);
@@ -244,7 +251,6 @@ public:
    */
   virtual already_AddRefed<nsIRadioGroupContainer> GetRadioGroupContainer();
 
-
 protected:
   // Helper method
   nsresult SetValueInternal(const nsAString& aValue,
@@ -275,8 +281,8 @@ protected:
   /**
    * Called when an attribute has just been changed
    */
-  void AfterSetAttr(PRInt32 aNameSpaceID, nsIAtom* aName,
-                    const nsAString* aValue, PRBool aNotify);
+  virtual void AfterSetAttr(PRInt32 aNameSpaceID, nsIAtom* aName,
+                            const nsAString* aValue, PRBool aNotify);
 
   void SelectAll(nsPresContext* aPresContext);
   PRBool IsImage() const
@@ -333,6 +339,17 @@ protected:
    * The current value of the input if it has been changed from the deafault
    */
   char*                    mValue;
+  /**
+   * The value of the input if it is a file input. This is the filename used
+   * when uploading a file. It is vital that this is kept separate from mValue
+   * so that it won't be possible to 'leak' the value from a text-input to a
+   * file-input. Additionally, the logic for this value is kept as simple as
+   * possible to avoid accidental errors where the wrong filename is used.
+   * Therefor the filename is always owned by this member, never by the frame.
+   * Whenever the frame wants to change the filename it has to call
+   * SetFileName to update this member.
+   */
+  nsAutoPtr<nsString>      mFileName;
 };
 
 //
@@ -371,10 +388,12 @@ NS_HTML_CONTENT_INTERFACE_MAP_BEGIN(nsHTMLInputElement,
   NS_INTERFACE_MAP_ENTRY(nsIDOMHTMLInputElement)
   NS_INTERFACE_MAP_ENTRY(nsIDOMNSHTMLInputElement)
   NS_INTERFACE_MAP_ENTRY(nsITextControlElement)
+  NS_INTERFACE_MAP_ENTRY(nsIFileControlElement)
   NS_INTERFACE_MAP_ENTRY(nsIRadioControlElement)
   NS_INTERFACE_MAP_ENTRY(nsIPhonetic)
   NS_INTERFACE_MAP_ENTRY(imgIDecoderObserver)
   NS_INTERFACE_MAP_ENTRY(nsIImageLoadingContent)
+  NS_INTERFACE_MAP_ENTRY(nsIDOMNSEditableElement)
   NS_INTERFACE_MAP_ENTRY_CONTENT_CLASSINFO(HTMLInputElement)
 NS_HTML_CONTENT_INTERFACE_MAP_END
 
@@ -398,7 +417,6 @@ nsHTMLInputElement::CloneNode(PRBool aDeep, nsIDOMNode** aReturn)
   switch (mType) {
     case NS_FORM_INPUT_TEXT:
     case NS_FORM_INPUT_PASSWORD:
-    case NS_FORM_INPUT_FILE:
       if (GET_BOOLBIT(mBitField, BF_VALUE_CHANGED)) {
         // We don't have our default value anymore.  Set our value on
         // the clone.
@@ -408,6 +426,11 @@ nsHTMLInputElement::CloneNode(PRBool aDeep, nsIDOMNode** aReturn)
         it->SetValueInternal(value, nsnull);
       }
       break;
+    case NS_FORM_INPUT_FILE:
+      if (mFileName) {
+        it->mFileName = new nsString(*mFileName);
+      }
+      break;
     case NS_FORM_INPUT_RADIO:
     case NS_FORM_INPUT_CHECKBOX:
       if (GET_BOOLBIT(mBitField, BF_CHECKED_CHANGED)) {
@@ -460,6 +483,8 @@ nsHTMLInputElement::AfterSetAttr(PRInt32 aNameSpaceID, nsIAtom* aName,
                                  const nsAString* aValue,
                                  PRBool aNotify)
 {
+  nsGenericHTMLFormElement::AfterSetAttr(aNameSpaceID, aName, aValue, aNotify);
+
   if (aNameSpaceID != kNameSpaceID_None) {
     return;
   }
@@ -602,8 +627,7 @@ nsHTMLInputElement::SetSize(PRUint32 aValue)
 NS_IMETHODIMP 
 nsHTMLInputElement::GetValue(nsAString& aValue)
 {
-  if (mType == NS_FORM_INPUT_TEXT || mType == NS_FORM_INPUT_PASSWORD ||
-      mType == NS_FORM_INPUT_FILE) {
+  if (mType == NS_FORM_INPUT_TEXT || mType == NS_FORM_INPUT_PASSWORD) {
     // No need to flush here, if there's no frame created for this
     // input yet, there won't be a value in it (that we don't already
     // have) even if we force it to be created
@@ -635,6 +659,17 @@ nsHTMLInputElement::GetValue(nsAString& aValue)
     return NS_OK;
   }
 
+  if (mType == NS_FORM_INPUT_FILE) {
+    if (mFileName) {
+      aValue = *mFileName;
+    }
+    else {
+      aValue.Truncate();
+    }
+
+    return NS_OK;
+  }
+  
   // Treat value == defaultValue for other input elements
   nsresult rv = GetAttr(kNameSpaceID_None, nsHTMLAtoms::value, aValue);
 
@@ -654,22 +689,28 @@ nsHTMLInputElement::SetValue(const nsAString& aValue)
 {
   // check security.  Note that setting the value to the empty string is always
   // OK and gives pages a way to clear a file input if necessary.
-  if (mType == NS_FORM_INPUT_FILE && !aValue.IsEmpty()) {
-    nsIScriptSecurityManager *securityManager =
-      nsContentUtils::GetSecurityManager();
+  if (mType == NS_FORM_INPUT_FILE) {
+    if (!aValue.IsEmpty()) {
+      nsIScriptSecurityManager *securityManager =
+        nsContentUtils::GetSecurityManager();
 
-    PRBool enabled;
-    nsresult rv =
-      securityManager->IsCapabilityEnabled("UniversalFileRead", &enabled);
-    NS_ENSURE_SUCCESS(rv, rv);
+      PRBool enabled;
+      nsresult rv =
+        securityManager->IsCapabilityEnabled("UniversalFileRead", &enabled);
+      NS_ENSURE_SUCCESS(rv, rv);
 
-    if (!enabled) {
-      // setting the value of a "FILE" input widget requires the
-      // UniversalFileRead privilege
-      return NS_ERROR_DOM_SECURITY_ERR;
+      if (!enabled) {
+        // setting the value of a "FILE" input widget requires the
+        // UniversalFileRead privilege
+        return NS_ERROR_DOM_SECURITY_ERR;
+      }
     }
+    SetFileName(aValue, PR_TRUE);
   }
-  SetValueInternal(aValue, nsnull);
+  else {
+    SetValueInternal(aValue, nsnull);
+  }
+
   return NS_OK;
 }
 
@@ -684,12 +725,42 @@ nsHTMLInputElement::TakeTextFrameValue(const nsAString& aValue)
   return NS_OK;
 }
 
+void
+nsHTMLInputElement::GetFileName(nsAString& aValue)
+{
+  if (mFileName) {
+    aValue = *mFileName;
+  }
+  else {
+    aValue.Truncate();
+  }
+}
+
+void
+nsHTMLInputElement::SetFileName(const nsAString& aValue, PRBool aUpdateFrame)
+{
+  // No big deal if |new| fails, we simply won't submit the file
+  mFileName = aValue.IsEmpty() ? nsnull : new nsString(aValue);
+
+  SetValueChanged(PR_TRUE);
+
+  if (aUpdateFrame) {
+    nsIFormControlFrame* formControlFrame = GetFormControlFrame(PR_FALSE);
+    if (formControlFrame) {
+      nsCOMPtr<nsPresContext> presContext = GetPresContext();
+      formControlFrame->SetProperty(presContext, nsHTMLAtoms::filename, aValue);
+    }
+  }
+}
+
 nsresult
 nsHTMLInputElement::SetValueInternal(const nsAString& aValue,
                                      nsITextControlFrame* aFrame)
 {
-  if (mType == NS_FORM_INPUT_TEXT || mType == NS_FORM_INPUT_PASSWORD ||
-      mType == NS_FORM_INPUT_FILE) {
+  NS_PRECONDITION(mType != NS_FORM_INPUT_FILE,
+                  "Don't call SetValueInternal for file inputs");
+
+  if (mType == NS_FORM_INPUT_TEXT || mType == NS_FORM_INPUT_PASSWORD) {
 
     nsITextControlFrame* textControlFrame = aFrame;
     nsIFormControlFrame* formControlFrame = textControlFrame;
@@ -707,9 +778,6 @@ nsHTMLInputElement::SetValueInternal(const nsAString& aValue,
     // File frames always own the value (if the frame is there).
     // Text frames have a bit that says whether they own the value.
     PRBool frameOwnsValue = PR_FALSE;
-    if (mType == NS_FORM_INPUT_FILE && formControlFrame) {
-      frameOwnsValue = PR_TRUE;
-    }
     if (textControlFrame) {
       textControlFrame->OwnsValue(&frameOwnsValue);
     }
@@ -731,6 +799,10 @@ nsHTMLInputElement::SetValueInternal(const nsAString& aValue,
     return mValue ? NS_OK : NS_ERROR_OUT_OF_MEMORY;
   }
 
+  if (mType == NS_FORM_INPUT_FILE) {
+    return NS_ERROR_UNEXPECTED;
+  }
+
   // If the value of a hidden input was changed, we mark it changed so that we
   // will know we need to save / restore the value.  Yes, we are overloading
   // the meaning of ValueChanged just a teensy bit to save a measly byte of
@@ -1293,6 +1365,21 @@ nsHTMLInputElement::HandleDOMEvent(nsPresContext* aPresContext,
     }
   }
 
+  // Don't allow mutation events which are targeted somewhere inside
+  // <input>, except if they are dispatched to the element itself.
+  if (!(NS_EVENT_FLAG_INIT & aFlags) &&
+      aEvent->eventStructType == NS_MUTATION_EVENT) {
+    return NS_OK;
+  }
+
+  if (!(NS_EVENT_FLAG_INIT & aFlags) &&
+      aEvent->message == NS_FORM_INPUT &&
+      formControlFrame) {
+    nsAutoString fName;
+    formControlFrame->GetProperty(nsHTMLAtoms::filename, fName);
+    mFileName = fName.IsEmpty() ? nsnull : new nsString(fName);
+  }
+
   //
   // Web pages expect the value of a radio button or checkbox to be set
   // *before* onclick and DOMActivate fire, and they expect that if they set
@@ -1771,15 +1858,23 @@ nsHTMLInputElement::ParseAttribute(nsIAtom* aAttribute,
       return PR_FALSE;
     }
 
-    mType = aResult.GetEnumValue();
-    if (mType == NS_FORM_INPUT_FILE) {
-      // If the type is being changed to file, set the element value
-      // to the empty string. This is for security.
-      // Call SetValueInternal so that this doesn't accidentally get caught
-      // in the security checks in SetValue.
+    // Make sure to do the check for newType being NS_FORM_INPUT_FILE and the
+    // corresponding SetValueInternal() call _before_ we set mType.  That way
+    // the logic in SetValueInternal() will work right (that logic makes
+    // assumptions about our frame based on mType, but we won't have had time
+    // to recreate frames yet -- that happens later in the SetAttr()
+    // process).
+    PRInt8 newType = aResult.GetEnumValue();
+    if (newType == NS_FORM_INPUT_FILE) {      
+      // These calls aren't strictly needed any more since we'll never
+      // confuse values and filenames. However they're there for backwards
+      // compat.
+      SetFileName(EmptyString(), PR_TRUE);
       SetValueInternal(EmptyString(), nsnull);
     }
 
+    mType = newType;
+    
     return PR_TRUE;
   }
   if (aAttribute == nsHTMLAtoms::width) {
@@ -2111,7 +2206,7 @@ nsHTMLInputElement::Reset()
     case NS_FORM_INPUT_FILE:
     {
       // Resetting it to blank should not perform security check
-      rv = SetValueInternal(EmptyString(), nsnull);
+      SetFileName(EmptyString(), PR_TRUE);
       break;
     }
     // Value is the same as defaultValue for hidden inputs
@@ -2232,22 +2327,6 @@ nsHTMLInputElement::SubmitNamesValues(nsIFormSubmission* aFormSubmission,
     return rv;
   }
 
-  // Get the value
-  nsAutoString value;
-  rv = GetValue(value);
-  if (NS_FAILED(rv)) {
-    return rv;
-  }
-
-  if (mType == NS_FORM_INPUT_SUBMIT && value.IsEmpty() &&
-      !HasAttr(kNameSpaceID_None, nsHTMLAtoms::value)) {
-    // Get our default value, which is the same as our default label
-    nsXPIDLString defaultValue;
-    nsContentUtils::GetLocalizedString(nsContentUtils::eFORMS_PROPERTIES,
-                                       "Submit", defaultValue);
-    value = defaultValue;
-  }
-      
   //
   // Submit file if it's input type=file and this encoding method accepts files
   //
@@ -2257,22 +2336,24 @@ nsHTMLInputElement::SubmitNamesValues(nsIFormSubmission* aFormSubmission,
     //
     nsCOMPtr<nsIFile> file;
  
-    if (StringBeginsWith(value, NS_LITERAL_STRING("file:"),
-                         nsCaseInsensitiveStringComparator())) {
-      // Converts the URL string into the corresponding nsIFile if possible.
-      // A local file will be created if the URL string begins with file://.
-      rv = NS_GetFileFromURLSpec(NS_ConvertUCS2toUTF8(value),
-                                 getter_AddRefs(file));
+    if (mFileName) {
+      if (StringBeginsWith(*mFileName, NS_LITERAL_STRING("file:"),
+                           nsCaseInsensitiveStringComparator())) {
+        // Converts the URL string into the corresponding nsIFile if possible.
+        // A local file will be created if the URL string begins with file://.
+        rv = NS_GetFileFromURLSpec(NS_ConvertUCS2toUTF8(*mFileName),
+                                   getter_AddRefs(file));
+      }
+
+      if (!file) {
+        // this is no "file://", try as local file
+        nsCOMPtr<nsILocalFile> localFile;
+        rv = NS_NewLocalFile(*mFileName, PR_FALSE, getter_AddRefs(localFile));
+        file = localFile;
+      }
     }
 
-    if (!file) {
-      // this is no "file://", try as local file
-      nsCOMPtr<nsILocalFile> localFile;
-      rv = NS_NewLocalFile(value, PR_FALSE, getter_AddRefs(localFile));
-      file = localFile;
-    }
-
-    if (NS_SUCCEEDED(rv)) {
+    if (file) {
 
       //
       // Get the leaf path name (to be submitted as the value)
@@ -2342,12 +2423,28 @@ nsHTMLInputElement::SubmitNamesValues(nsIFormSubmission* aFormSubmission,
     // If we can't even make a truncated filename, submit empty string
     // rather than sending everything
     //
-    aFormSubmission->AddNameFilePair(this, name, value,
+    aFormSubmission->AddNameFilePair(this, name, EmptyString(),
         nsnull, NS_LITERAL_CSTRING("application/octet-stream"),
         PR_FALSE);
     return rv;
   }
 
+  // Get the value
+  nsAutoString value;
+  rv = GetValue(value);
+  if (NS_FAILED(rv)) {
+    return rv;
+  }
+
+  if (mType == NS_FORM_INPUT_SUBMIT && value.IsEmpty() &&
+      !HasAttr(kNameSpaceID_None, nsHTMLAtoms::value)) {
+    // Get our default value, which is the same as our default label
+    nsXPIDLString defaultValue;
+    nsContentUtils::GetLocalizedString(nsContentUtils::eFORMS_PROPERTIES,
+                                       "Submit", defaultValue);
+    value = defaultValue;
+  }
+      
   // Submit
   // (for type=image, only submit if value is non-null)
   if (mType != NS_FORM_INPUT_IMAGE || !value.IsEmpty()) {
@@ -2395,7 +2492,6 @@ nsHTMLInputElement::SaveState()
     case NS_FORM_INPUT_PASSWORD:
       break;
     case NS_FORM_INPUT_TEXT:
-    case NS_FORM_INPUT_FILE:
     case NS_FORM_INPUT_HIDDEN:
       {
         if (GET_BOOLBIT(mBitField, BF_VALUE_CHANGED)) {
@@ -2414,6 +2510,17 @@ nsHTMLInputElement::SaveState()
         }
         break;
       }
+    case NS_FORM_INPUT_FILE:
+      {
+        if (mFileName) {
+          rv = GetPrimaryPresState(this, &state);
+          if (state) {
+            rv = state->SetStateProperty(NS_LITERAL_STRING("f"), *mFileName);
+            NS_ASSERTION(NS_SUCCEEDED(rv), "value save failed!");
+          }
+        }
+        break;
+      }
   }
   
   if (GET_BOOLBIT(mBitField, BF_DISABLED_CHANGED)) {
@@ -2503,7 +2610,6 @@ nsHTMLInputElement::RestoreState(nsPresState* aState)
       }
 
     case NS_FORM_INPUT_TEXT:
-    case NS_FORM_INPUT_FILE:
     case NS_FORM_INPUT_HIDDEN:
       {
         nsAutoString value;
@@ -2514,6 +2620,17 @@ nsHTMLInputElement::RestoreState(nsPresState* aState)
         }
         break;
       }
+    case NS_FORM_INPUT_FILE:
+      {
+        nsAutoString value;
+        rv = aState->GetStateProperty(NS_LITERAL_STRING("f"), value);
+        NS_ASSERTION(NS_SUCCEEDED(rv), "value restore failed!");
+        if (rv == NS_STATE_PROPERTY_EXISTS) {
+          // Can there really be a frame at this point?
+          SetFileName(value, PR_TRUE);
+        }
+        break;
+      }
   }
   
   nsAutoString disabled;
