@@ -100,6 +100,7 @@ nsCaret::nsCaret()
 , mLastContentOffset(0)
 , mLastHint(nsIFrameSelection::HINTLEFT)
 #ifdef IBMBIDI
+, mLastBidiLevel(0)
 , mKeyboardRTL(PR_FALSE)
 #endif
 {
@@ -231,6 +232,13 @@ NS_IMETHODIMP nsCaret::SetCaretDOMSelection(nsISelection *aDOMSel)
 {
   NS_ENSURE_ARG_POINTER(aDOMSel);
   mDomSelectionWeak = do_GetWeakReference(aDOMSel);   // weak reference to pres shell
+  if (mVisible)
+  {
+    // Stop the caret from blinking in its previous location.
+    StopBlinking();
+    // Start the caret blinking in the new location.
+    StartBlinking();
+  }
   return NS_OK;
 }
 
@@ -534,22 +542,27 @@ nsCaret::DrawAtPositionWithHint(nsIDOMNode*             aNode,
   nsPresContext *presContext = presShell->GetPresContext();
   if (presContext && presContext->BidiEnabled())
   {
-    presShell->GetCaretBidiLevel(&bidiLevel);
-    if (bidiLevel & BIDI_LEVEL_UNDEFINED)
-    {
-      PRUint8 newBidiLevel;
-      bidiLevel &= ~BIDI_LEVEL_UNDEFINED;
-      // There has been a reflow, so we reset the cursor Bidi level to the level of the current frame
-      if (!presContext) // Use the style default or default to 0
+    if (mDrawn) {
+      bidiLevel = mLastBidiLevel;
+    } else {
+      presShell->GetCaretBidiLevel(&bidiLevel);
+      if (bidiLevel & BIDI_LEVEL_UNDEFINED)
       {
-        newBidiLevel = theFrame->GetStyleVisibility()->mDirection;
-      }
-      else
-      {
-        newBidiLevel = NS_GET_EMBEDDING_LEVEL(theFrame);
-        presShell->SetCaretBidiLevel(newBidiLevel);
-        bidiLevel = newBidiLevel;
+        PRUint8 newBidiLevel;
+        bidiLevel &= ~BIDI_LEVEL_UNDEFINED;
+        // There has been a reflow, so we reset the cursor Bidi level to the level of the current frame
+        if (!presContext) // Use the style default or default to 0
+        {
+          newBidiLevel = theFrame->GetStyleVisibility()->mDirection;
+        }
+        else
+        {
+          newBidiLevel = NS_GET_EMBEDDING_LEVEL(theFrame);
+          presShell->SetCaretBidiLevel(newBidiLevel);
+          bidiLevel = newBidiLevel;
+        }
       }
+      mLastBidiLevel = bidiLevel;
     }
 
     PRInt32 start;
@@ -1066,6 +1079,9 @@ void nsCaret::GetCaretRectAndInvert(nsIFrame* aFrame, PRInt32 aFrameOffset)
           if (NS_SUCCEEDED(domSelection->SelectionLanguageChange(mKeyboardRTL)))
           {
             mRendContext->PopState();
+#ifdef DONT_REUSE_RENDERING_CONTEXT
+            mRendContext = nsnull;
+#endif
             return;
           }
         }
