@@ -820,9 +820,9 @@ nsresult nsMsgSearchTerm::MatchBody (nsIMsgSearchScopeTerm *scope, PRUint32 offs
   // since we don't have a MIME parser handy, and we want to err on the
   // side of too many hits rather than not enough, we'll assume in that
   // general direction. Blech. ### FIX ME 
-  // bug fix #88935: for stateful csids like JIS, we don't want to decode
-  // quoted printable since it contains '='.
-  PRBool isQuotedPrintable =  /*!(mail_csid & STATEFUL) && */
+  // bug fix #314637: for stateful charsets like ISO-2022-JP, we don't
+  // want to decode quoted printable since it contains '='.
+  PRBool isQuotedPrintable = !nsMsgI18Nstateful_charset(folderCharset) &&
     (PL_strchr (m_value.string, '=') == nsnull);
   
   while (!endOfFile && result == boolContinueLoop)
@@ -1782,97 +1782,6 @@ int nsMsgResultElement::CompareByFolderInfoPtrs (const void *e1, const void *e2)
 int nsMsgResultElement::Compare (const void *e1, const void *e2)
 {
   int ret = 0;
-#ifdef HAVE_SEARCH_PORT
-  // Bad karma to cast away const, but they're my objects anway.
-  // Maybe if we go through and const everything this should be a const ptr.
-  nsMsgResultElement *re1 = *(nsMsgResultElement**) e1;
-  nsMsgResultElement *re2 = *(nsMsgResultElement**) e2;
-  
-  NS_ASSERTION(re1->IsValid(), "invalid result element1 in resultElement::Compare");
-  NS_ASSERTION(re2->IsValid(), "invalid result element2 in resultElement::Compare");
-  
-  nsMsgSearchAttribute attrib = re1->m_adapter->m_scope->m_frame->m_sortAttribute;
-  
-  const nsMsgSearchValue *v1 = re1->GetValueRef (attrib);
-  const nsMsgSearchValue *v2 = re2->GetValueRef (attrib);
-  
-  if (!v1 || !v2)
-    return ret; // search result doesn't contain the attrib we want to sort on
-  
-  switch (attrib)
-  {
-  case nsMsgSearchAttrib::Date:
-    {
-      // on Win 3.1, the requisite 'int' return type is a short, so use a 
-      // real time_t for comparison
-      time_t date = v1->u.date - v2->u.date;
-      if (date)
-        ret = ((long)date) < 0 ? -1 : 1;
-      else
-        ret = 0;
-    }
-    break;
-  case nsMsgSearchAttrib::Priority:
-    ret = v1->u.priority - v2->u.priority;
-    break;
-  case nsMsgSearchAttrib::Size:
-    ret = v1->u.size - v2->u.size;
-    break;
-  case nsMsgSearchAttrib::MsgStatus:
-    {
-      // Here's a totally arbitrary sorting protocol for msg status
-      uint32 s1, s2;
-      
-      s1 = v1->u.msgStatus & ~MSG_FLAG_REPLIED;
-      s2 = v2->u.msgStatus & ~MSG_FLAG_REPLIED;
-      if (s1 || s2)
-        ret = s1 - s2;
-      else
-      {
-        s1 = v1->u.msgStatus & ~MSG_FLAG_FORWARDED;
-        s2 = v2->u.msgStatus & ~MSG_FLAG_FORWARDED;
-        if (s1 || s2)
-          ret = s1 - s2;
-        else
-        {
-          s1 = v1->u.msgStatus & ~MSG_FLAG_READ;
-          s2 = v2->u.msgStatus & ~MSG_FLAG_READ;
-          if (s1 || s2)
-            ret = s1 - s2;
-          else
-            // statuses don't contain any flags we're interested in, 
-            // so they're equal as far as we care
-            ret = 0;
-        }
-      }
-    }
-    break;
-  default:
-    if (attrib == nsMsgSearchAttrib::Subject)
-    {
-      // Special case for subjects, so "Re:foo" sorts under 'f' not 'r'
-      const char *s1 = v1->u.string;
-      const char *s2 = v2->u.string;
-      NS_MsgStripRE (&s1, NULL);
-      NS_MsgStripRE (&s2, NULL);
-      ret = PL_strcasecomp (s1, s2);
-    }
-    else
-      ret = strcasecomp (v1->u.string, v2->u.string);
-  }
-  // ### need different hack for this.
-  // qsort's default sort order is ascending, so in order to get descending
-  // behavior, we'll tell qsort a lie and reverse the comparison order.
-  if (re1->m_adapter->m_scope->m_frame->m_descending && ret != 0)
-    if (ret < 0)
-      ret = 1;
-    else
-      ret = -1;
-    
-    // <0 --> e1 less than e2
-    // 0  --> e1 equal to e2
-    // >0 --> e1 greater than e2
-#endif
     return ret;
 }
 
