@@ -62,6 +62,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <limits.h>
+#include <errno.h>
 
 #if defined(XP_WIN)
 # include <windows.h>
@@ -72,7 +73,10 @@
 # define R_OK 04
 # define access _access
 # define snprintf _snprintf
+# define putenv _putenv
 # define fchmod(a,b)
+# define mkdir(path, perm) _mkdir(path)
+# define chdir(path) _chdir(path)
 #else
 # include <sys/wait.h>
 # include <unistd.h>
@@ -265,6 +269,46 @@ static const char kWhitespace[] = " \t";
 static const char kNL[] = "\r\n";
 static const char kQuote[] = "\"";
 
+//-----------------------------------------------------------------------------
+// LOGGING
+
+static FILE *gLogFP = NULL;
+
+static void LogInit()
+{
+  if (gLogFP)
+    return;
+
+  char logFile[MAXPATHLEN];
+  snprintf(logFile, MAXPATHLEN, "%s/update.log", gSourcePath);
+
+  gLogFP = fopen(logFile, "w");
+}
+
+static void LogFinish()
+{
+  if (!gLogFP)
+    return;
+
+  fclose(gLogFP);
+  gLogFP = NULL;
+}
+
+static void LogPrintf(const char *fmt, ... )
+{
+  if (!gLogFP)
+    return;
+
+  va_list ap;
+  va_start(ap, fmt);
+  vfprintf(gLogFP, fmt, ap);
+  va_end(ap);
+}
+
+#define LOG(args) LogPrintf args
+
+//-----------------------------------------------------------------------------
+
 static inline PRUint32
 mmin(PRUint32 a, PRUint32 b)
 {
@@ -313,19 +357,25 @@ mstrtok(const char *delims, char **str)
 // Ensure that the directory containing this file exists.
 static int ensure_parent_dir(const char *path)
 {
+  int rv = OK;
+
   char *slash = (char *) strrchr(path, '/');
   if (slash)
   {
     *slash = '\0';
-    ensure_parent_dir(path);
-#ifdef XP_WIN
-    _mkdir(path);
-#else
-    mkdir(path, 0755);
-#endif
+    rv = ensure_parent_dir(path);
+    if (rv == OK) {
+      rv = mkdir(path, 0755);
+      // If the directory already exists, then ignore the error.
+      if (rv < 0 && errno != EEXIST) {
+        rv = WRITE_ERROR;
+      } else {
+        rv = OK;
+      }
+    }
     *slash = '/';
   }
-  return OK;
+  return rv;
 }
 
 static int copy_file(const char *spath, const char *dpath)
@@ -337,12 +387,16 @@ static int copy_file(const char *spath, const char *dpath)
   struct stat ss;
 
   AutoFD sfd = open(spath, O_RDONLY | _O_BINARY);
-  if (sfd < 0 || fstat(sfd, &ss))
-    return IO_ERROR;
+  if (sfd < 0 || fstat(sfd, &ss)) {
+    LOG(("copy_file: failed to open or stat: %d,%s,%d\n", (int) sfd, spath, errno));
+    return READ_ERROR;
+  }
 
   AutoFD dfd = open(dpath, O_WRONLY | O_TRUNC | O_CREAT | _O_BINARY, ss.st_mode);
-  if (dfd < 0)
-    return IO_ERROR;
+  if (dfd < 0) {
+    LOG(("copy_file: failed to open: %s,%d\n", dpath, errno));
+    return WRITE_ERROR;
+  }
 
   char buf[BUFSIZ];
   int sc;
@@ -354,53 +408,19 @@ static int copy_file(const char *spath, const char *dpath)
         break;
       bp += dc;
     }
-    if (dc < 0)
-      return IO_ERROR;
+    if (dc < 0) {
+      LOG(("copy_file: failed to write: %d\n", errno));
+      return WRITE_ERROR;
+    }
+  }
+  if (sc < 0) {
+    LOG(("copy_file: failed to read: %d\n", errno));
+    return READ_ERROR;
   }
-  if (sc < 0)
-    return IO_ERROR;
 
   return OK;
 }
 
-//-----------------------------------------------------------------------------
-// LOGGING
-
-static FILE *gLogFP = NULL;
-
-static void LogInit()
-{
-  if (gLogFP)
-    return;
-
-  char logFile[MAXPATHLEN];
-  snprintf(logFile, MAXPATHLEN, "%s/update.log", gSourcePath);
-
-  gLogFP = fopen(logFile, "w");
-}
-
-static void LogFinish()
-{
-  if (!gLogFP)
-    return;
-
-  fclose(gLogFP);
-  gLogFP = NULL;
-}
-
-static void LogPrintf(const char *fmt, ... )
-{
-  if (!gLogFP)
-    return;
-
-  va_list ap;
-  va_start(ap, fmt);
-  vfprintf(gLogFP, fmt, ap);
-  va_end(ap);
-}
-
-#define LOG(args) LogPrintf args
-
 //-----------------------------------------------------------------------------
 
 #define BACKUP_EXT ".moz-backup"
@@ -411,11 +431,7 @@ static int backup_create(const char *path)
   char backup[MAXPATHLEN];
   snprintf(backup, sizeof(backup), "%s" BACKUP_EXT, path);
 
-  int rv = copy_file(path, backup);
-  if (rv)
-    return IO_ERROR;
-
-  return OK;
+  return copy_file(path, backup);
 }
 
 // Copy the backup copy of the specified file back overtop
@@ -428,11 +444,11 @@ static int backup_restore(const char *path)
 
   int rv = copy_file(backup, path);
   if (rv)
-    return IO_ERROR;
+    return rv;
 
   rv = remove(backup);
   if (rv)
-    return IO_ERROR;
+    return WRITE_ERROR;
 
   return OK;
 }
@@ -445,7 +461,7 @@ static int backup_discard(const char *path)
 
   int rv = remove(backup);
   if (rv)
-    return IO_ERROR;
+    return WRITE_ERROR;
 
   return OK;
 }
@@ -527,7 +543,7 @@ RemoveFile::Prepare()
   // We expect the file to exist if we are to remove it.
   int rv = access(mFile, F_OK);
   if (rv) {
-    LOG(("file cannot be removed because it does not exist; skipping"));
+    LOG(("file cannot be removed because it does not exist; skipping\n"));
     mSkip = 1;
     return OK;
   }
@@ -541,8 +557,10 @@ RemoveFile::Prepare()
     rv = access(".", W_OK);
   }
 
-  if (rv)
-    return IO_ERROR;
+  if (rv) {
+    LOG(("access failed: %d\n", errno));
+    return WRITE_ERROR;
+  }
 
   return OK;
 }
@@ -555,16 +573,30 @@ RemoveFile::Execute()
   if (mSkip)
     return OK;
 
+  // We expect the file to exist if we are to remove it.  We check here as well
+  // as in PREPARE since we might have been asked to remove the same file more
+  // than once: bug 311099.
+  int rv = access(mFile, F_OK);
+  if (rv) {
+    LOG(("file cannot be removed because it does not exist; skipping\n"));
+    mSkip = 1;
+    return OK;
+  }
+
   // save a complete copy of the old file, and then remove the
   // old file.  we'll clean up the copy in Finish.
 
-  int rv = backup_create(mFile);
-  if (rv)
+  rv = backup_create(mFile);
+  if (rv) {
+    LOG(("backup_create failed: %d\n", rv));
     return rv;
+  }
 
   rv = remove(mFile);
-  if (rv)
-    return IO_ERROR;
+  if (rv) {
+    LOG(("remove failed: %d\n", rv));
+    return WRITE_ERROR;
+  }
 
   return OK;
 }
@@ -630,7 +662,7 @@ AddFile::Execute()
 
     rv = remove(mFile);
     if (rv)
-      return IO_ERROR;
+      return WRITE_ERROR;
   }
   else
   {
@@ -695,27 +727,27 @@ PatchFile::LoadSourceFile(int ofd)
   struct stat os;
   int rv = fstat(ofd, &os);
   if (rv)
-    return IO_ERROR;
+    return READ_ERROR;
 
   if (PRUint32(os.st_size) != header.slen)
-    return BSP_ERROR_CORRUPT;
+    return UNEXPECTED_ERROR;
 
   buf = (unsigned char*) malloc(header.slen);
   if (!buf)
-    return BSP_ERROR_NOMEM;
+    return MEM_ERROR;
 
   int r = header.slen;
   unsigned char *rb = buf;
   while (r) {
     int c = read(ofd, rb, mmin(BUFSIZ,r));
     if (c < 0)
-      return IO_ERROR;
+      return READ_ERROR;
 
     r -= c;
     rb += c;
 
     if (c == 0 && r)
-      return BSP_ERROR_CORRUPT;
+      return UNEXPECTED_ERROR;
   }
 
   // Verify that the contents of the source file correspond to what we expect.
@@ -724,7 +756,7 @@ PatchFile::LoadSourceFile(int ofd)
 
   if (crc != header.scrc32) {
     LOG(("CRC check failed\n"));
-    return BSP_ERROR_CORRUPT;
+    return CRC_ERROR;
   }
   
   return OK;
@@ -774,7 +806,7 @@ PatchFile::Prepare()
 
   pfd = open(spath, O_RDONLY | _O_BINARY);
   if (pfd < 0)
-    return IO_ERROR;
+    return READ_ERROR;
 
   rv = MBS_ReadHeader(pfd, &header);
   if (rv)
@@ -782,7 +814,7 @@ PatchFile::Prepare()
 
   AutoFD ofd = open(mFile, O_RDONLY | _O_BINARY);
   if (ofd < 0)
-    return IO_ERROR;
+    return READ_ERROR;
 
   rv = LoadSourceFile(ofd);
   if (rv)
@@ -799,7 +831,7 @@ PatchFile::Execute()
 
   struct stat ss;
   if (stat(mFile, &ss))
-    return IO_ERROR;
+    return READ_ERROR;
 
   int rv = backup_create(mFile);
   if (rv)
@@ -807,11 +839,11 @@ PatchFile::Execute()
 
   rv = remove(mFile);
   if (rv)
-    return IO_ERROR;
+    return WRITE_ERROR;
 
   AutoFD ofd = open(mFile, O_WRONLY | O_TRUNC | O_CREAT | _O_BINARY, ss.st_mode);
   if (ofd < 0)
-    return IO_ERROR;
+    return WRITE_ERROR;
 
   return MBS_ApplyPatch(&header, pfd, buf, ofd);
 }
@@ -948,21 +980,24 @@ PatchIfFile::Finish(int status)
 
 //-----------------------------------------------------------------------------
 
+#ifdef XP_WIN
+#include "nsWindowsRestart.cpp"
+#endif
+
 static void
-LaunchCallbackApp(int argc, char **argv)
+LaunchCallbackApp(const char *workingDir, int argc, char **argv)
 {
+  putenv("NO_EM_RESTART=");
+
+  // Run from the specified working directory (see bug 312360).
+  chdir(workingDir);
+
 #if defined(USE_EXECV)
   execv(argv[0], argv);
 #elif defined(XP_MACOSX)
   LaunchChild(argc, argv);
 #elif defined(XP_WIN)
-  // _spawnv gets angry if the executable path has spaces in it.
-  char shortPath[_MAX_PATH];
-  ::GetShortPathName(argv[0], shortPath, sizeof(shortPath));
-
-  argv[0] = shortPath;
-
-  _spawnv(_P_NOWAIT, shortPath, argv);
+  WinLaunchChild(argv[0], argc, argv);
 #else
 # warning "Need implementaton of LaunchCallbackApp"
 #endif
@@ -980,7 +1015,15 @@ WriteStatusFile(int status)
   if (fd < 0)
     return;
 
-  const char *text = (status == OK) ? "succeeded\n" : "failed\n";
+  const char *text;
+
+  char buf[32];
+  if (status == OK) {
+    text = "succeeded\n";
+  } else {
+    snprintf(buf, sizeof(buf), "failed: %d\n", status);
+    text = buf;
+  }
   write(fd, text, strlen(text));
 }
 
@@ -1020,7 +1063,7 @@ int main(int argc, char **argv)
   // be altered.
 
   if (argc < 3) {
-    fprintf(stderr, "Usage: updater <dir-path> <parent-pid> [callback args...]\n");
+    fprintf(stderr, "Usage: updater <dir-path> <parent-pid> [working-dir callback args...]\n");
     return 1;
   }
 
@@ -1048,6 +1091,19 @@ int main(int argc, char **argv)
 
   gSourcePath = argv[1];
 
+#ifdef XP_WIN
+  // By opening a file handle to the executable, we can tell the OS to
+  // prevent new application processes from launching while we are
+  // updating.
+
+  HANDLE exefile = NULL;
+
+  if (argc > 5)
+    exefile = CreateFile(argv[4], DELETE | GENERIC_WRITE,
+                         0, // no sharing!
+                         NULL, OPEN_EXISTING, 0, NULL);
+#endif
+
   LogInit();
 
   // Run update process on a background thread.  ShowProgressUI may return
@@ -1060,10 +1116,16 @@ int main(int argc, char **argv)
 
   LogFinish();
 
+#ifdef XP_WIN
+  if (exefile)
+    CloseHandle(exefile);
+#endif
+
   // The callback to execute is given as the last N arguments of our command
-  // line.
-  if (argc > 3)
-    LaunchCallbackApp(argc - 3, argv + 3);
+  // line.  The first of those arguments specifies the working directory for
+  // the callback.
+  if (argc > 4)
+    LaunchCallbackApp(argv[3], argc - 4, argv + 4);
 
   return 0;
 }
@@ -1171,29 +1233,29 @@ int DoUpdate()
 
   AutoFD mfd = open(manifest, O_RDONLY | _O_BINARY);
   if (mfd < 0)
-    return -1;
+    return READ_ERROR;
 
   struct stat ms;
   rv = fstat(mfd, &ms);
   if (rv)
-    return IO_ERROR;
+    return READ_ERROR;
 
   char *mbuf = (char*) malloc(ms.st_size + 1);
   if (!mbuf)
-    return BSP_ERROR_NOMEM;
+    return MEM_ERROR;
 
   int r = ms.st_size;
   char *rb = mbuf;
   while (r) {
     int c = read(mfd, rb, mmin(SSIZE_MAX,r));
     if (c < 0)
-      return IO_ERROR;
+      return READ_ERROR;
 
     r -= c;
     rb += c;
 
     if (c == 0 && r)
-      return BSP_ERROR_CORRUPT;
+      return UNEXPECTED_ERROR;
   }
   mbuf[ms.st_size] = '\0';
 
@@ -1231,7 +1293,7 @@ int DoUpdate()
     }
 
     if (!action)
-      return BSP_ERROR_NOMEM;
+      return MEM_ERROR;
 
     rv = action->Parse(line);
     if (rv)
