@@ -24,6 +24,7 @@
  *   Benjamin Smedberg <benjamin@smedbergs.us>
  *   Ben Goodger <ben@mozilla.org>
  *   Fredrik Holmqvist <thesuckiestemail@yahoo.se>
+ *   Ben Turner <mozilla@songbirdnest.com>
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -132,7 +133,6 @@
 #endif //XP_BEOS
 
 #ifdef XP_WIN
-#include <windows.h>
 #include <process.h>
 #endif
 
@@ -240,22 +240,6 @@ extern "C" {
   extern void InstallUnixSignalHandlers(const char *ProgramName);
 #endif
 
-#if defined(XP_WIN)
-/**
- * Return a malloc'd copy of the given string surrounded with double quotes.
- */
-static char *NewQuotedString(const char *input)
-{
-  int len = strlen(input);
-  char *p = (char *) malloc(len + 3);
-  p[0] = '\"';
-  memcpy(&p[1], input, len);
-  p[len + 1] = '\"';
-  p[len + 2] = '\0';
-  return p;
-}
-#endif
-
 int    gArgc;
 char **gArgv;
 
@@ -1090,6 +1074,7 @@ XRE_GetBinaryPath(const char* argv0, nsILocalFile* *aResult)
 #elif defined(XP_UNIX)
   struct stat fileStat;
   char exePath[MAXPATHLEN];
+  char tmpPath[MAXPATHLEN];
 
   rv = NS_ERROR_FAILURE;
 
@@ -1132,8 +1117,8 @@ XRE_GetBinaryPath(const char* argv0, nsILocalFile* *aResult)
     char *newStr = pathdup;
     char *token;
     while ( (token = nsCRT::strtok(newStr, ":", &newStr)) ) {
-      sprintf(exePath, "%s/%s", token, argv0);
-      if (stat(exePath, &fileStat) == 0) {
+      sprintf(tmpPath, "%s/%s", token, argv0);
+      if (realpath(tmpPath, exePath) && stat(exePath, &fileStat) == 0) {
         found = PR_TRUE;
         break;
       }
@@ -1182,6 +1167,10 @@ XRE_GetBinaryPath(const char* argv0, nsILocalFile* *aResult)
 
 #define NS_ERROR_LAUNCHED_CHILD_PROCESS NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_PROFILE, 200)
 
+#ifdef XP_WIN
+#include "nsWindowsRestart.cpp"
+#endif
+
 // If aBlankCommandLine is true, then the application will be launched with a
 // blank command line instead of being launched with the same command line that
 // it was initially started with.
@@ -1212,15 +1201,7 @@ static nsresult LaunchChild(nsINativeAppSupport* aNative,
     return rv;
 
 #if defined(XP_WIN)
-  // We must shorten the path to a 8.3 path, since that's all _execv can
-  // handle (otherwise segments after any spaces that might exist in the path
-  // will be converted into parameters, and really weird things will happen)
-  char shortPath[MAXPATHLEN];
-  ::GetShortPathName(exePath.get(), shortPath, MAXPATHLEN);
-
-  gRestartArgv[0] = shortPath;
-
-  if (_execv(shortPath, gRestartArgv) == -1)
+  if (!WinLaunchChild(exePath.get(), gRestartArgc, gRestartArgv))
     return NS_ERROR_FAILURE;
 #elif defined(XP_OS2)
   if (_execv(exePath.get(), gRestartArgv) == -1)
@@ -1474,7 +1455,7 @@ SelectProfile(nsIProfileLock* *aResult, nsINativeAppSupport* aNative,
     const char *dummy;
     CheckArg("p", &dummy);
     CheckArg("profile", &dummy);
-    CheckArg("profilemanager", &dummy);
+    CheckArg("profilemanager");
 
     return NS_LockProfilePath(lf, localDir, nsnull, aResult);
   }
@@ -1648,7 +1629,8 @@ SelectProfile(nsIProfileLock* *aResult, nsINativeAppSupport* aNative,
 
 static PRBool
 CheckCompatibility(nsIFile* aProfileDir, const nsCString& aVersion,
-                   nsIFile* aXULRunnerDir, nsIFile* aAppDir)
+                   const nsCString& aOSABI, nsIFile* aXULRunnerDir,
+                   nsIFile* aAppDir)
 {
   nsCOMPtr<nsIFile> file;
   aProfileDir->Clone(getter_AddRefs(file));
@@ -1670,6 +1652,13 @@ CheckCompatibility(nsIFile* aProfileDir, const nsCString& aVersion,
   if (!aVersion.Equals(buf))
     return PR_FALSE;
 
+  rv = parser.GetString("Compatibility", "LastOSABI", buf);
+  if (NS_FAILED(rv))
+    return PR_FALSE;
+
+  if (!aOSABI.Equals(buf))
+    return PR_FALSE;
+
   rv = parser.GetString("Compatibility", "LastPlatformDir", buf);
   if (NS_FAILED(rv))
     return PR_FALSE;
@@ -1714,7 +1703,8 @@ static void BuildVersion(nsCString &aBuf)
 
 static void
 WriteVersion(nsIFile* aProfileDir, const nsCString& aVersion,
-             nsIFile* aXULRunnerDir, nsIFile* aAppDir)
+             const nsCString& aOSABI, nsIFile* aXULRunnerDir,
+             nsIFile* aAppDir)
 {
   nsCOMPtr<nsIFile> file;
   aProfileDir->Clone(getter_AddRefs(file));
@@ -1744,6 +1734,10 @@ WriteVersion(nsIFile* aProfileDir, const nsCString& aVersion,
   PR_Write(fd, kHeader, sizeof(kHeader) - 1);
   PR_Write(fd, aVersion.get(), aVersion.Length());
 
+  static const char kOSABIHeader[] = NS_LINEBREAK "LastOSABI=";
+  PR_Write(fd, kOSABIHeader, sizeof(kOSABIHeader) - 1);
+  PR_Write(fd, aOSABI.get(), aOSABI.Length());
+
   static const char kPlatformDirHeader[] = NS_LINEBREAK "LastPlatformDir=";
 
   PR_Write(fd, kPlatformDirHeader, sizeof(kPlatformDirHeader) - 1);
@@ -1921,13 +1915,6 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
 
   int i;
   for (i = 0; i < argc; ++i) {
-#if defined(XP_WIN)
-    if (strchr(argv[i], ' ')) {
-      gRestartArgv[i] = NewQuotedString(argv[i]);
-      if (!gRestartArgv[i]) return 1;
-    }
-    else
-#endif
     gRestartArgv[i] = argv[i];
   }
   gRestartArgv[argc] = nsnull;
@@ -2081,8 +2068,29 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
 #endif
 
 #if defined(MOZ_UPDATER)
+  // If this is a XULRunner app then the updater needs to know the base
+  // directory that contains the application.ini file. This should be the
+  // parent of the xulrunner directory on Windows/Linux and it should be the
+  // Contents directory on MacOSX. Just in case someone packaged their app
+  // incorrectly we'll pass the directory here.
+  nsCOMPtr<nsIFile> greDir = dirProvider.GetAppDir();
+  NS_ENSURE_TRUE(greDir, 1);
+
+  nsCOMPtr<nsIFile> appDir;
+  PRBool dummy;
+  rv = dirProvider.GetFile("resource:app",
+                           &dummy,
+                           getter_AddRefs(appDir));
+  if (NS_FAILED(rv)) {
+    // This must not be a XULRunner app
+    appDir = greDir;
+  }
+
   // Check for and process any available updates
-  ProcessUpdates(dirProvider.GetAppDir(), gRestartArgc, gRestartArgv);
+  ProcessUpdates(greDir,
+                 appDir,
+                 gRestartArgc,
+                 gRestartArgv);
 #endif
 
   nsCOMPtr<nsIProfileLock> profileLock;
@@ -2111,10 +2119,17 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
   nsCAutoString version;
   BuildVersion(version);
 
+#ifdef TARGET_OS_ABI
+    NS_NAMED_LITERAL_CSTRING(osABI, TARGET_OS_ABI);
+#else
+    // No TARGET_XPCOM_ABI, but at least the OS is known
+    NS_NAMED_LITERAL_CSTRING(osABI, OS_TARGET "_UNKNOWN");
+#endif
+
   // Check for version compatibility with the last version of the app this 
   // profile was started with.  The format of the version stamp is defined
   // by the BuildVersion function.
-  PRBool versionOK = CheckCompatibility(profD, version,
+  PRBool versionOK = CheckCompatibility(profD, version, osABI,
                                         dirProvider.GetAppDir(),
                                         gAppData->directory);
 
@@ -2127,7 +2142,7 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
   //
   if (gSafeMode) {
     RemoveComponentRegistries(profD, profLD, PR_FALSE);
-    WriteVersion(profD, NS_LITERAL_CSTRING("Safe Mode"),
+    WriteVersion(profD, NS_LITERAL_CSTRING("Safe Mode"), osABI,
                  dirProvider.GetAppDir(), gAppData->directory);
   }
   else if (versionOK) {
@@ -2152,7 +2167,7 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
     upgraded = PR_TRUE;
 
     // Write out version
-    WriteVersion(profD, version,
+    WriteVersion(profD, version, osABI,
                  dirProvider.GetAppDir(), gAppData->directory);
   }
 
@@ -2196,6 +2211,24 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
       // So we can open and close windows during startup
       appStartup->EnterLastWindowClosingSurvivalArea();
 
+      if (gDoMigration) {
+        nsCOMPtr<nsIFile> file;
+        dirProvider.GetAppDir()->Clone(getter_AddRefs(file));
+        file->AppendNative(NS_LITERAL_CSTRING("override.ini"));
+        nsINIParser parser;
+        nsCOMPtr<nsILocalFile> localFile(do_QueryInterface(file));
+        nsresult rv = parser.Init(localFile);
+        if (NS_SUCCEEDED(rv)) {
+          nsCAutoString buf;
+          rv = parser.GetString("XRE", "EnableProfileMigrator", buf);
+          if (NS_SUCCEEDED(rv)) {
+            if (buf[0] == '0' || buf[0] == 'f' || buf[0] == 'F') {
+              gDoMigration = PR_FALSE;
+            }
+          }
+        }
+      }
+
       // Profile Migration
       if (gAppData->flags & NS_XRE_ENABLE_PROFILE_MIGRATOR && gDoMigration) {
         gDoMigration = PR_FALSE;
@@ -2295,6 +2328,10 @@ XRE_main(int argc, char* argv[], const nsXREAppData* aAppData)
 
         UpdatePrebinding();
 #endif
+        nsCOMPtr<nsIObserverService> obsService
+          (do_GetService("@mozilla.org/observer-service;1"));
+        if (obsService)
+          obsService->NotifyObservers(nsnull, "final-ui-startup", nsnull);        
 
         rv = cmdLine->Run();
         NS_ENSURE_SUCCESS_LOG(rv, 1);
